/**************************************************************************
     picpuz.cpp
     Free software licensed under GNU General Public License v.2
***************************************************************************/

#include "zlib.h"

#define gtitle "picpuz.06  picture puzzle"                                 //  v.06
#define nodither GDK_RGB_DITHER_NONE,0,0
#define Tindex(row,col) row * Tcols + col                                  //  map row/col to linear index
#define drand(seed,range) (drandz(&seed) * range)                          //  random double, 0.0 to 0.9999 * range
#define lrand(seed,range) (lrandz(&seed) % range)                          //  random integer, 0 to range-1
#define interp GDK_INTERP_HYPER                                            //  image scaling interpolation
#define solidline GDK_LINE_SOLID,GDK_CAP_BUTT,GDK_JOIN_MITER               //  solid line attributes
#define dottedline GDK_LINE_DOUBLE_DASH,GDK_CAP_BUTT,GDK_JOIN_MITER        //  dotted line attributes

GtkWidget      *win1, *vbox1, *dwin1, *stbar1;                             //  main and drawing window
GtkWidget      *win2 = 0, *dwin2;                                          //  reference and drawing window
GdkGC          *gdkgc = 0;                                                 //  graphics context
GdkPixbuf      *iPixbuf = 0;                                               //  image pixbuff
GdkPixmap      *iPixmap = 0;                                               //  image pixmap
GdkColor       black, white;
GdkColormap    *colormap = 0;
uint           maxcolor = 0xffff;
uint           stbarctx;
int8           dashes[2] = { 3, 3 };

timeval  rtime;
int64    rseed;                                                            //  random number seed

char     iFile[MAXFCC] = "";                                               //  image file pathname
char     pname[100];                                                       //  puzzle name
int      winW = 800, winH = 600;                                           //  window size
int      imageW, imageH;                                                   //  image size
int      tileU = 80;                                                       //  tile size, user setpoint
int      tileW, tileH;                                                     //  actual tile size used
int      Trows, Tcols;                                                     //  puzzle tile rows, cols
int      Ntiles = 0;                                                       //  puzzle tile count
int      Nmoves = 0;                                                       //  tiles moved count
int      Nhome = 0;                                                        //  tiles at home position
int      Mstate = 0;                                                       //  mouse, tile select state

struct t_tileposn {
   int      row, col;                                                      //  map tile position
};
t_tileposn   *wposn = 0;                                                   //  window position of home tile
t_tileposn   *hposn = 0;                                                   //  home position of window tile

void winpaint();                                                           //  window paint function
void menufunc(GtkMenuItem *item, char *menu);                              //  menu processor
void new_puzzle();                                                         //  create new puzzle
void retile_puzzle();                                                      //  re-tile puzzle
void mixup_puzzle();                                                       //  mix-up pizzle tiles
void show_ref_image();                                                     //  display reference image
void save_puzzle();                                                        //  save puzzle for later
void resume_puzzle();                                                      //  resume saved puzzle
void move_tiles(int N);                                                    //  move some tiles home
void change_borders();                                                     //  change tile borderlines
void quit();                                                               //  exit application
void destroyfunc();                                                        //  window destroy signal function
int  puzzle_status();                                                      //  test puzzle status
void help();                                                               //  display help file

void init_puzzle(int init);                                                //  initialize puzzle
void tile_window(int init);                                                //  paint tiles to window
void mouse_event(GtkWidget *, GdkEventButton *);                           //  handle mouse clicks
void swap_tiles(int row1, int col1, int row2, int col2);                   //  swap two tiles
void draw_tile(int row, int col);                                          //  draw tile at window position
void stbar_update();                                                       //  update status bar


//  main program

int main(int argc, char *argv[])
{
   gtk_init(&argc, &argv);
   zlockInit();
   
   gettimeofday(&rtime,0);                                                 //  random number seed
   rseed = rtime.tv_sec;

   win1 = gtk_window_new(GTK_WINDOW_TOPLEVEL);                             //  create main window
   gtk_window_set_title(GTK_WINDOW(win1),gtitle);
   gtk_window_set_position(GTK_WINDOW(win1),GTK_WIN_POS_CENTER);
   gtk_window_set_default_size(GTK_WINDOW(win1),winW,winH);
   
   vbox1 = gtk_vbox_new(0,0);                                              //  add vertical packing box
   gtk_container_add(GTK_CONTAINER(win1),vbox1);

   create_toolbar(toolbar)                                                 //  add toolbar and buttons
   gtk_box_pack_start(GTK_BOX(vbox1),toolbar,0,0,0);
   add_toolbar_stock_item(toolbar,tbnew,"new\npuzzle","create new puzzle from image file",OPEN,menufunc)
   add_toolbar_stock_item(toolbar,tbtile,"retile\npuzzle","make tiles bigger/smaller",PREFERENCES,menufunc)
   add_toolbar_stock_item(toolbar,tbmix,"mix\ntiles","randomize tiles",REFRESH,menufunc)
   add_toolbar_stock_item(toolbar,tbshow,"show\nimage","view image file",INFO,menufunc)
   add_toolbar_stock_item(toolbar,tbsave,"save\npuzzle","save puzzle to file",SAVE,menufunc)
   add_toolbar_stock_item(toolbar,tbres,"resume\npuzzle","reload saved puzzle",OPEN,menufunc)
   add_toolbar_stock_item(toolbar,tbdo1,"do 1\ntile","move one random tile home",EXECUTE,menufunc)
   add_toolbar_stock_item(toolbar,tbdoN,"do 10\ntiles","move ten random tiles home",EXECUTE,menufunc)
   add_toolbar_stock_item(toolbar,tbline,"change\nborders","change tile borderlines",PROPERTIES,menufunc)
   add_toolbar_stock_item(toolbar,tbquit,"quit","quit picpuz",QUIT,menufunc)
   add_toolbar_stock_item(toolbar,tbhelp,"help","view help document",HELP,menufunc)
   
   dwin1 = gtk_drawing_area_new();                                         //  add drawing window
   gtk_box_pack_start(GTK_BOX(vbox1),dwin1,1,1,0);
   
   stbar1 = gtk_statusbar_new();                                           //  add status bar
   gtk_box_pack_start(GTK_BOX(vbox1),stbar1,0,0,0);
   stbarctx = gtk_statusbar_get_context_id(GTK_STATUSBAR(stbar1),"all");
   stbar_update();

   gtk_widget_add_events(dwin1,GDK_BUTTON_PRESS_MASK);                     //  connect signals
   gtk_widget_add_events(dwin1,GDK_BUTTON_RELEASE_MASK);
   G_SIGNAL(dwin1,"button-press-event",mouse_event,0)
   G_SIGNAL(dwin1,"button-release-event",mouse_event,0)
   G_SIGNAL(dwin1,"expose-event",winpaint,0)
   G_SIGNAL(win1,"destroy",destroyfunc,0)

   gtk_widget_show_all(win1);                                              //  show all widgets

   gdkgc = gdk_gc_new(dwin1->window);                                      //  initz. graphics context
   black.red = black.green = black.blue = 0;
   white.red = white.green = white.blue = maxcolor;
   colormap = gtk_widget_get_colormap(dwin1);
   gdk_rgb_find_color(colormap,&black);
   gdk_rgb_find_color(colormap,&white);
   gdk_gc_set_foreground(gdkgc,&black);
   gdk_gc_set_background(gdkgc,&white);
   gdk_gc_set_dashes(gdkgc,0,(gint8 *) dashes,2);
   gdk_gc_set_fill(gdkgc,GDK_SOLID);
   gdk_gc_set_line_attributes(gdkgc,1,solidline);

   gtk_main();                                                             //  process window events
   return 0;
}


//  paint window when created, exposed, or resized

void winpaint()
{
   gdk_window_clear(dwin1->window);
   if (Ntiles) tile_window(0);                                             //  paint tiles
   return;
}


//  process button selection events

void menufunc(GtkMenuItem *item, char *menu)
{
   if (strEqu(menu,"new\npuzzle")) new_puzzle();
   if (strEqu(menu,"retile\npuzzle")) retile_puzzle();
   if (strEqu(menu,"mix\ntiles")) mixup_puzzle();
   if (strEqu(menu,"show\nimage")) show_ref_image();
   if (strEqu(menu,"save\npuzzle")) save_puzzle();
   if (strEqu(menu,"resume\npuzzle")) resume_puzzle();
   if (strEqu(menu,"do 1\ntile")) move_tiles(1);
   if (strEqu(menu,"do 10\ntiles")) move_tiles(10);
   if (strEqu(menu,"change\nborders")) change_borders();
   if (strEqu(menu,"quit")) quit();
   if (strEqu(menu,"help")) help();
}


//  create a new puzzle from an image file

void new_puzzle()
{
   char        *file;
   
   if (puzzle_status()) return;                                            //  do not discard

   file = zfile_open("load image file",iFile);
   if (! file) return;

   gdk_window_clear(dwin1->window);                                        //  set no puzzle
   Ntiles = Nmoves = Nhome = 0;

   strncpy0(iFile,file,MAXFCC-1);                                          //  initialize puzzle
   init_puzzle(1);
   g_free(file);
   return;
}


//  re-tile puzzle according to user input

void retile_puzzle()
{
   zdialog     *zd;
   int         stat, nn;
   
   if (! Ntiles) return;
   if (puzzle_status()) return;                                            //  do not discard

   zd = zdialog_new("set new tile size","OK","cancel",0);                  //  dialog for tile size
   zdialog_add_widget(zd,"hbox","hb1","dialog");
   zdialog_add_widget(zd,"label","lb1","hb1","tile size (pixels)");
   zdialog_add_widget(zd,"spin","spin","hb1","20,200,1,80");
   zdialog_stuff(zd,"spin",tileU);

   stat = zdialog_run(zd);
   zdialog_destroy(zd);
   if (stat != 1) return;                                                  //  cancel

   zdialog_fetch(zd,"spin",nn);
   if (nn == tileU) return;                                                //  no change

   tileU = nn;                                                             //  new setpoint tile size
   tile_window(1);                                                         //  initialize puzzle
   return;
}


//  mix-up the tiles

void mixup_puzzle()
{
   int      row1, col1, row2, col2, ii;

   if (puzzle_status()) return;                                            //  do not discard
   
   for (row1 = 0; row1 < Trows; row1++)                                    //  randomize tile positions
   for (col1 = 0; col1 < Tcols; col1++)
   {
      ii = Tindex(row1,col1);
      if (wposn[ii].row != row1) continue;
      if (wposn[ii].col != col1) continue;
      row2 = lrand(rseed,Trows);
      col2 = lrand(rseed,Tcols);
      swap_tiles(row1,col1,row2,col2);
      zsleep(0.01);
      zmainloop();
   }
   
   stbar_update();                                                         //  update status bar
   return;
}


//  display a reference image in a new window

void show_ref_image()
{
   void win2_paint();
   void win2_destroy();

   if (! *iFile) return;   

   if (win2) {
      win2_paint();
      return;
   }

   win2 = gtk_window_new(GTK_WINDOW_TOPLEVEL);                             //  create window
   gtk_window_set_position(GTK_WINDOW(win2),GTK_WIN_POS_CENTER);
   gtk_window_set_default_size(GTK_WINDOW(win2),400,300);
   gtk_window_set_transient_for(GTK_WINDOW(win2),GTK_WINDOW(win1));
   
   dwin2 = gtk_drawing_area_new();                                         //  add drawing window
   gtk_container_add(GTK_CONTAINER(win2),dwin2);

   G_SIGNAL(win2,"destroy",win2_destroy,0)                                 //  connect events
   G_SIGNAL(dwin2,"expose-event",win2_paint,0)

   gtk_widget_show_all(win2);                                              //  show all widgets
   return;
}

void win2_paint()
{
   GdkPixbuf   *refPixbuf;
   int         x, y, winx, winy;

   if (! iPixbuf) return;

   gtk_window_set_title(GTK_WINDOW(win2),pname);
   gdk_window_clear(dwin2->window);

   winx = dwin2->allocation.width;                                         //  curr. window size
   winy = dwin2->allocation.height;
   
   x = int(1.0 * winy * imageW / imageH);                                  //  preserve X/Y ratio
   if (winx > x) winx = x;
   y = int(1.0 * winx * imageH / imageW);
   if (winy > y) winy = y;
   
   refPixbuf = gdk_pixbuf_scale_simple(iPixbuf,winx,winy,interp);          //  scale image to window

   gdk_draw_pixbuf(dwin2->window,0,refPixbuf,0,0,0,0,-1,-1,nodither);      //  paint window
   g_object_unref(refPixbuf);
   return;
}

void win2_destroy()
{
   win2 = 0;
   return;
}


//  save puzzle to a file for later resumption

void save_puzzle()
{
   FILE           *fid;
   char           *file;
   int            ii, row, col;
   
   if (! Ntiles) return;
   
   strncatv(pname,99,".puz",0);

   file = zfile_save("save puzzle to a file",pname);
   if (! file) return;

   fid = fopen(file,"w");
   if (! fid) { 
      wpauseOK("cannot open: %s",file);
      g_free(file);
      return;
   }

   fprintf(fid,"%s \n",iFile);                                             //  save image file
   fprintf(fid," %d %d \n",Ntiles,Nhome);                                  //  save tile count and no. home
   fprintf(fid," %d %d \n",Trows,Tcols);                                   //  save row and col counts
   
   for (row = 0; row < Trows; row++) {                                     //  save tile position data
      for (col = 0; col < Tcols; col++) {
         ii = Tindex(row,col);
         fprintf(fid," %d,%d ",wposn[ii].row, wposn[ii].col);
      }
      fprintf(fid,"\n");
   }
   
   fclose(fid);
   g_free(file);
   
   Nmoves = 0;                                                             //  reset move count
   return;
}


//  resume a saved puzzle

void resume_puzzle()
{
   FILE        *fid;
   int         stat, row1, col1, row2, col2, ii;
   char        *file, *pp;
   
   if (puzzle_status()) return;                                            //  do not discard
   
   gdk_window_clear(dwin1->window);                                        //  set no puzzle
   Ntiles = Nmoves = Nhome = 0;
   stbar_update();
   *iFile = 0;
   
   file = zfile_open("load puzzle from a file",0);
   if (! file) return;

   fid = fopen(file,"r");
   if (! fid) {
      wpauseOK("cannot open: %s",file);
      g_free(file);
      return;
   }

   pp = fgets_trim(iFile,MAXFCC,fid,1);                                    //  read image file name
   if (! pp) goto badfile;
   
   stat = fscanf(fid," %d %d ",&Ntiles,&Nhome);
   if (stat != 2) goto badfile;
   
   stat = fscanf(fid," %d %d ",&Trows,&Tcols);                             //  read row and col counts
   if (stat != 2) goto badfile;
   if (Ntiles != Trows * Tcols) goto badfile;

   if (wposn) delete [] wposn;
   wposn = new t_tileposn[Ntiles];   
   
   for (row1 = 0; row1 < Trows; row1++) {                                  //  read tile position data
      for (col1 = 0; col1 < Tcols; col1++) {
         stat = fscanf(fid," %d,%d ",&row2,&col2);
         if (stat != 2) goto badfile;
         if (row2 < 0 || row2 >= Trows) goto badfile;
         if (col2 < 0 || col2 >= Tcols) goto badfile;
         ii = Tindex(row1,col1);
         wposn[ii].row = row2;
         wposn[ii].col = col2;
      }
   }
   
   fclose(fid);
   g_free(file);

   init_puzzle(0);                                                         //  initialize, preserve tile data
   return;
   
badfile:
   fclose(fid);
   g_free(file);
   wpauseOK("file format is wrong");
   *iFile = 0;                                                             //  set no puzzle
   Ntiles = Nmoves = Nhome = 0;
   stbar_update();
   return;
}


//  move tiles into place automatically

void move_tiles(int nn1)
{
   int      row1, col1, row2, col2, ii, nn2;
   
   if (! Ntiles) return;
   
   while (true)
   {
      row1 = lrand(rseed,Trows);                                           //  random starting tile
      col1 = lrand(rseed,Tcols);
      nn2 = Ntiles;
      
      while (true)                                                         //  scan for tile to move
      {
         ii = Tindex(row1,col1);
         row2 = wposn[ii].row;                                             //  curr. position
         col2 = wposn[ii].col;
         if (row1 != row2 || col1 != col2) break;                          //  not at home position
         if (--nn2 == 0) goto mdone;
         if (++col1 < Tcols) continue;                                     //  look at next tile
         col1 = 0;
         if (++row1 < Trows) continue;
         row1 = 0;
      }

      swap_tiles(row1,col1,row2,col2);                                     //  move tile home
      zsleep(0.02);
      zmainloop();
      if (--nn1 == 0) goto mdone;
   }

mdone:
   stbar_update();                                                         //  update status bar
   return;
}


//  chage tile borderline style                                            v.05

void change_borders()
{
   static int  linestyle = 0;

   linestyle++;
   if (linestyle > 2) linestyle = 0;

   switch (linestyle) {
   
   case 0: 
      gdk_gc_set_foreground(gdkgc,&black);
      gdk_gc_set_line_attributes(gdkgc,1,solidline);
      break;

   case 1: 
      gdk_gc_set_foreground(gdkgc,&white);
      gdk_gc_set_line_attributes(gdkgc,1,solidline);
      break;

   case 2: 
      gdk_gc_set_foreground(gdkgc,&black);
      gdk_gc_set_line_attributes(gdkgc,1,dottedline);
      break;
   }

   winpaint();
   return;
}


//  quit function with last chance to save unfinished puzzle

void quit()
{
   if (puzzle_status()) return;                                            //  do not discard
   gtk_main_quit();   
   return;
}


//  signal function for window destroy signal

void destroyfunc()
{
   if (puzzle_status()) save_puzzle();                                     //  do not discard
   gtk_main_quit();   
   return;
}


//  test puzzle before discarding, give user a chance to save
//  returns  0 = completed, unchanged, or user says discard
//           1 = incomplete, modified, and user says keep

int puzzle_status()
{
   if (! Ntiles) return 0;                                                 //  no puzzle
   if (! Nmoves) return 0;                                                 //  no new moves
   if (Nhome == Ntiles) return 0;                                          //  puzzle done
   if (wpauseYN("DISCARD unfinished puzzle?")) return 0;                   //  OK to discard
   return 1;                                                               //  keep
}


//  thread function to display help file

void help()                                                                //  menu function
{
   pthread_t   tid;
   void * help_thread(void *);
   pthread_create(&tid,0,help_thread,0);                                   //  launch thread function
   return;
}

void * help_thread(void *)                                                 //  thread function
{
   int      cc, ii, err;
   char     *pp, helppath[MAXFCC+20], command[MAXFCC+50];
   char     *trycomm[3] = { PDF_VIEWERS };
   
   cc = readlink("/proc/self/exe",helppath,MAXFCC);                        //  get execution file directory
   if (cc > 0) helppath[cc] = 0;
   else strcpy(helppath,"/picpuz/picpuz.x");                               //  can't - supply dummy
   pp = strrchr(helppath,'/');                                             //  isolate directory only
   if (pp) *pp = 0;

   for (ii = 0; ii < 3; ii++) 
   {
      sprintf(command,"%s %s/picpuz-guide.pdf",trycomm[ii],helppath);      //  command /helppath/picpuz-guide.pdf
      err = system(command);
      if (! err) break;
   }

   return 0;
}


//  initialize puzzle from image file with optional preservation 
//  of existing tile data (row and column counts, tile positions)
//  called by new_puzzle() and resume_puzzle()

void init_puzzle(int newp)
{
   GError      **gerror = 0;
   char        *pp;
   
   if (! *iFile) return;

   pp = strrchr(iFile,'/');                                                //  puzzle name = image file name
   if (! pp++) pp = iFile;
   strncpy0(pname,pp,99);

   if (iPixbuf) g_object_unref(iPixbuf);
   iPixbuf = gdk_pixbuf_new_from_file(iFile,gerror);                       //  load file into pixbuf
   if (!iPixbuf) {
      wpauseOK("image type not recognized:\n %s",iFile);
      *iFile = 0;
      return;
   }

   imageW = gdk_pixbuf_get_width(iPixbuf);                                 //  get image size
   imageH = gdk_pixbuf_get_height(iPixbuf);

   if (imageW < 300 || imageH < 300) {
      wpauseOK("image too small, please select another");
      g_object_unref(iPixbuf);
      iPixbuf = 0;
      *iFile = 0;
      return;
   }

   tile_window(newp);                                                      //  paint tiles on window
   if (win2) win2_paint();                                                 //  paint reference image
   return;
}


//  create tile pixmaps and paint tiles on window
//  called by init_puzzle(), retile_puzzle(), winpaint()

void tile_window(int newp)
{
   GdkPixbuf   *wPixbuf;
   int         winW, winH, x1, y1, ii, row, col;
   
   gdk_window_clear(dwin1->window);                                        //  black window

   if (! iPixbuf) return;

   winW = dwin1->allocation.width;                                         //  curr. window size
   winH = dwin1->allocation.height;
   
   winW = winW - 4;                                                        //  to keep margins visible
   winH = winH - 4;
   
   x1 = int(1.0 * winH * imageW / imageH);                                 //  preserve image X/Y ratio
   if (winW > x1) winW = x1;
   y1 = int(1.0 * winW * imageH / imageW);
   if (winH > y1) winH = y1;

   if (newp)                                                               //  new puzzle
   {
      Tcols = int(1.0 * winW / tileU + 0.5);                               //  best fit to user tile size
      Trows = int(1.0 * winH / tileU + 0.5);
      Ntiles = Trows * Tcols;
      Nhome = Ntiles;

      if (wposn) delete [] wposn;                                          //  new window position map
      wposn = new t_tileposn[Ntiles];   

      for (row = 0; row < Trows; row++)
      for (col = 0; col < Tcols; col++)
      {
         ii = Tindex(row,col);                                             //  all window positions = home
         wposn[ii].row = row;   
         wposn[ii].col = col;
      }
   }

   if (hposn) delete [] hposn;                                             //  new home position map
   hposn = new t_tileposn[Ntiles];   

   for (row = 0; row < Trows; row++)                                       //  initialize home position map
   for (col = 0; col < Tcols; col++)                                       //    from window position map
   {
      ii = Tindex(row,col);
      ii = Tindex(wposn[ii].row,wposn[ii].col);
      hposn[ii].row = row;   
      hposn[ii].col = col;
   }
   
   tileW = winW / Tcols;                                                   //  actual tile size to use
   tileH = winH / Trows;

   winW = Tcols * tileW;                                                   //  synch. window to tile size
   winH = Trows * tileH;

   wPixbuf = gdk_pixbuf_scale_simple(iPixbuf,winW,winH,interp);            //  scale image to window size

   if (iPixmap) g_object_unref(iPixmap);                                   //  allocate new pixmap
   iPixmap = gdk_pixmap_new(dwin1->window,winW,winH,-1);
   gdk_draw_pixbuf(iPixmap,0,wPixbuf,0,0,0,0,-1,-1,nodither);              //  copy scaled image to pixmap
   g_object_unref(wPixbuf);

   for (row = 0; row < Trows; row++)                                       //  draw tile pixmaps on window
   for (col = 0; col < Tcols; col++) 
      draw_tile(row,col);
   
   Mstate = 0;                                                             //  no tile selected
   stbar_update();                                                         //  update status bar
   return;
}


//  process mouse events (button down, button up)

void mouse_event(GtkWidget *, GdkEventButton *event)
{
   static int  row1, col1, row2, col2;
   int         button, ii, x, y;

   if (! Ntiles) return;

   button = event->button;                                                 //  1/2/3 = left/middle/right
   x = int(event->x);
   y = int(event->y);

   if (x < 1 || x >= (Tcols * tileW)) goto mret0;                          //  ignore if not within picture
   if (y < 1 || y >= (Trows * tileH)) goto mret0;

   if (button == 3) {                                                      //  if right button, 
      row1 = y / tileH;                                                    //    get tile belonging here
      col1 = x / tileW;
      ii = Tindex(row1,col1);
      row2 = wposn[ii].row;                                                //  where it is now
      col2 = wposn[ii].col;
      swap_tiles(row1,col1,row2,col2);                                     //  swap
      goto mret0;
   }

   if (Mstate < 2) {                                                       //  1st tile selection
      row1 = y / tileH;
      col1 = x / tileW;
   }

   else {                                                                  //  2nd tile selection
      row2 = y / tileH;
      col2 = x / tileW;
   }
   
   Mstate++;

   if (Mstate < 4) goto mret;
   swap_tiles(row1,col1,row2,col2);                                        //  last button up

mret0:
   Mstate = 0;

mret:
   stbar_update();                                                         //  update status bar
   return;
}


//  swap two tile positions on window

void swap_tiles(int row1, int col1, int row2, int col2)
{
   #define  SWAP(x,y) (temp = (x), (x) = (y), (y) = temp)

   int      ii1, ii2, temp;
   
   if (! Ntiles) return;
   
   ii1 = Tindex(row1,col1);                                                //  count tiles moved from home
   ii2 = Tindex(row2,col2);
   if (hposn[ii1].row == row1 && hposn[ii1].col == col1) Nhome--;
   if (hposn[ii2].row == row2 && hposn[ii2].col == col2) Nhome--;

   ii1 = Tindex(row1,col1);                                                //  swap home positions
   ii2 = Tindex(row2,col2);
   SWAP(hposn[ii1].row,hposn[ii2].row);
   SWAP(hposn[ii1].col,hposn[ii2].col);

   ii1 = Tindex(hposn[ii1].row,hposn[ii1].col);                            //  swap window positions
   ii2 = Tindex(hposn[ii2].row,hposn[ii2].col);
   SWAP(wposn[ii1].row,wposn[ii2].row);
   SWAP(wposn[ii1].col,wposn[ii2].col);
   
   draw_tile(row1,col1);                                                   //  draw tiles at new positions
   draw_tile(row2,col2);

   Nmoves++;                                                               //  incr. move count

   ii1 = Tindex(row1,col1);                                                //  count tiles moved to home
   ii2 = Tindex(row2,col2);
   if (hposn[ii1].row == row1 && hposn[ii1].col == col1) Nhome++;
   if (hposn[ii2].row == row2 && hposn[ii2].col == col2) Nhome++;

   return;
}


//  draw tile at window position (row, col)

void draw_tile(int row, int col)
{
   void     draw_body(int row, int col);
   void     draw_lobe(int row, int col);
   void     draw_hole(int row, int col);
   
   if (col > 0) {
      draw_body(row,col-1);                                                //  refresh tile to left
      draw_lobe(row,col-1);
      draw_hole(row,col-1);
   }

   draw_body(row,col);                                                     //  refresh this tile
   draw_lobe(row,col);
   draw_hole(row,col);

   if (col < Tcols-1) draw_lobe(row,col+1);                                //  refresh lobe from tile to right
   return;
}

//  draw tile body at window position (no lobes or holes)

void draw_body(int row2, int col2)
{
   int      ii, row1, col1;
   int      x1, y1, x2, y2;
   
   ii = Tindex(row2,col2);                                                 //  get home position
   row1 = hposn[ii].row;
   col1 = hposn[ii].col;
   
   x1 = col1 * tileW;                                                      //  draw tile pixmap to its
   y1 = row1 * tileH;                                                      //    current window position
   x2 = col2 * tileW;
   y2 = row2 * tileH;
   gdk_draw_drawable(dwin1->window,gdkgc,iPixmap,x1,y1,x2,y2,tileW,tileH);
   gdk_draw_rectangle(dwin1->window,gdkgc,0,x2,y2,tileW,tileH);
   
   return;
}

//  draw protruding lobe from left side of puzzle tile at window position

void draw_lobe(int row2, int col2)
{
   int      ii, row1, col1;
   int      x1, y1, x2, y2, px, py, pw, ph;
   int64    seed;
   
   if (col2 == 0) return;                                                  //  window position on left edge

   ii = Tindex(row2,col2);                                                 //  get home position for tile
   row1 = hposn[ii].row;
   col1 = hposn[ii].col;
   if (col1 == 0) return;                                                  //  home position on left edge
   
   seed = row1 + col1;
   px = int(0.2 * tileW + 0.5);
   py = int(drand(seed,0.6) * tileH + 0.1 * tileH);
   pw = int(0.2 * tileW + 0.5);
   ph = int(0.2 * tileH + 0.5);
   x1 = col1 * tileW - px;
   y1 = row1 * tileH + py;
   x2 = col2 * tileW - px;
   y2 = row2 * tileH + py;
   gdk_draw_drawable(dwin1->window,gdkgc,iPixmap,x1,y1,x2,y2,pw+1,ph);

   gdk_draw_line(dwin1->window,gdkgc,x2,y2,x2+pw,y2);
   gdk_draw_line(dwin1->window,gdkgc,x2,y2,x2,y2+ph);
   gdk_draw_line(dwin1->window,gdkgc,x2,y2+ph,x2+pw,y2+ph);

   return;
}

//  draw recess hole on right side where lobe from right fits in

void draw_hole(int row2, int col2)
{
   int      ii, row1, col1;
   int      x2, y2, px, py, pw, ph;
   int64    seed;
   
   ii = Tindex(row2,col2);                                                 //  get home position for tile
   row1 = hposn[ii].row;
   col1 = hposn[ii].col;
   if (++col1 == Tcols) return;                                            //  tile to the right, unless on edge

   seed = row1 + col1;
   px = int(0.2 * tileW + 0.5);
   py = int(drand(seed,0.6) * tileH + 0.1 * tileH);
   pw = int(0.2 * tileW + 0.5);
   ph = int(0.2 * tileH + 0.5);
   x2 = col2 * tileW + tileW - px;
   y2 = row2 * tileH + py;
   gdk_draw_rectangle(dwin1->window,gdkgc,1,x2,y2,pw,ph);
   
   return;
}   


//  update statistics on status bar

void stbar_update()
{
   char     stbarmessage[50];

   sprintf(stbarmessage," tiles home: %d/%d",Nhome,Ntiles);
   if (Mstate > 1) strcat(stbarmessage,"   1st tile selected");
   gtk_statusbar_pop(GTK_STATUSBAR(stbar1),stbarctx);
   gtk_statusbar_push(GTK_STATUSBAR(stbar1),stbarctx,stbarmessage);
   return;
}


